@@ -1,5 +1,5 @@
 //
-// Copyright (w) 2016-2017 Vinnie Falco (vinnie dot falco at gmail dot com)
+// Copyright (c) 2016-2019 Vinnie Falco (vinnie dot falco at gmail dot com)
 //
 // Distributed under the Boost Software License, Version 1.0. (See accompanying
 // file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
@@ -10,7 +10,7 @@
 // Test that header file is self-contained.
 #include <boost/beast/websocket/stream.hpp>
 
-#include <boost/asio/io_service.hpp>
+#include <boost/asio/io_context.hpp>
 #include <boost/asio/strand.hpp>
 
 #include "test.hpp"
@@ -26,8 +26,6 @@ public:
     void
     doTestWrite(Wrap const& w)
     {
-        using boost::asio::buffer;
-
         permessage_deflate pmd;
         pmd.client_enable = false;
         pmd.server_enable = false;
@@ -47,7 +45,7 @@ public:
             catch(system_error const& se)
             {
                 BEAST_EXPECTS(
-                    se.code() == boost::asio::error::operation_aborted,
+                    se.code() == net::error::operation_aborted,
                     se.code().message());
             }
         }
@@ -59,7 +57,7 @@ public:
             ws.auto_fragment(false);
             ws.binary(false);
             std::string const s = "Hello, world!";
-            w.write(ws, buffer(s));
+            w.write(ws, net::buffer(s));
             multi_buffer b;
             w.read(ws, b);
             BEAST_EXPECT(ws.got_text());
@@ -71,7 +69,7 @@ public:
         [&](ws_type_t<deflateSupported>& ws)
         {
             ws.text(true);
-            w.write(ws, boost::asio::const_buffer{});
+            w.write(ws, net::const_buffer{});
             multi_buffer b;
             w.read(ws, b);
             BEAST_EXPECT(ws.got_text());
@@ -85,8 +83,8 @@ public:
             ws.auto_fragment(false);
             ws.binary(false);
             std::string const s = "Hello, world!";
-            w.write_some(ws, false, buffer(s.data(), 5));
-            w.write_some(ws, true, buffer(s.data() + 5, s.size() - 5));
+            w.write_some(ws, false, net::buffer(s.data(), 5));
+            w.write_some(ws, true, net::buffer(s.data() + 5, s.size() - 5));
             multi_buffer b;
             w.read(ws, b);
             BEAST_EXPECT(ws.got_text());
@@ -101,8 +99,8 @@ public:
             std::size_t const chop = 3;
             BOOST_ASSERT(chop < s.size());
             w.write_some(ws, false,
-                buffer(s.data(), chop));
-            w.write_some(ws, true, buffer(
+                net::buffer(s.data(), chop));
+            w.write_some(ws, true, net::buffer(
                 s.data() + chop, s.size() - chop));
             flat_buffer b;
             w.read(ws, b);
@@ -115,7 +113,7 @@ public:
         {
             ws.auto_fragment(false);
             std::string const s = "Hello";
-            w.write(ws, buffer(s));
+            w.write(ws, net::buffer(s));
             flat_buffer b;
             w.read(ws, b);
             BEAST_EXPECT(buffers_to_string(b.data()) == s);
@@ -126,9 +124,9 @@ public:
         [&](ws_type_t<deflateSupported>& ws)
         {
             ws.auto_fragment(false);
-            ws.write_buffer_size(16);
+            ws.write_buffer_bytes(16);
             std::string const s(32, '*');
-            w.write(ws, buffer(s));
+            w.write(ws, net::buffer(s));
             flat_buffer b;
             w.read(ws, b);
             BEAST_EXPECT(buffers_to_string(b.data()) == s);
@@ -140,7 +138,7 @@ public:
         {
             ws.auto_fragment(true);
             std::string const s(16384, '*');
-            w.write(ws, buffer(s));
+            w.write(ws, net::buffer(s));
             flat_buffer b;
             w.read(ws, b);
             BEAST_EXPECT(buffers_to_string(b.data()) == s);
@@ -158,7 +156,7 @@ public:
                 w.accept(ws);
                 ws.auto_fragment(false);
                 std::string const s = "Hello";
-                w.write(ws, buffer(s));
+                w.write(ws, net::buffer(s));
                 flat_buffer b;
                 w.read(ws, b);
                 BEAST_EXPECT(buffers_to_string(b.data()) == s);
@@ -184,7 +182,7 @@ public:
                 w.accept(ws);
                 ws.auto_fragment(true);
                 std::string const s(16384, '*');
-                w.write(ws, buffer(s));
+                w.write(ws, net::buffer(s));
                 flat_buffer b;
                 w.read(ws, b);
                 BEAST_EXPECT(buffers_to_string(b.data()) == s);
@@ -203,8 +201,6 @@ public:
     void
     doTestWriteDeflate(Wrap const& w)
     {
-        using boost::asio::buffer;
-
         permessage_deflate pmd;
         pmd.client_enable = true;
         pmd.server_enable = true;
@@ -215,7 +211,7 @@ public:
         {
             auto const& s = random_string();
             ws.binary(true);
-            w.write(ws, buffer(s));
+            w.write(ws, net::buffer(s));
             flat_buffer b;
             w.read(ws, b);
             BEAST_EXPECT(buffers_to_string(b.data()) == s);
@@ -230,8 +226,8 @@ public:
             // This call should produce no
             // output due to compression latency.
             w.write_some(ws, false,
-                buffer(s.data(), chop));
-            w.write_some(ws, true, buffer(
+                net::buffer(s.data(), chop));
+            w.write_some(ws, true, net::buffer(
                 s.data() + chop, s.size() - chop));
             flat_buffer b;
             w.read(ws, b);
@@ -244,7 +240,7 @@ public:
         {
             auto const& s = random_string();
             ws.binary(true);
-            w.write(ws, buffer(s));
+            w.write(ws, net::buffer(s));
             flat_buffer b;
             w.read(ws, b);
             BEAST_EXPECT(buffers_to_string(b.data()) == s);
@@ -254,8 +250,6 @@ public:
     void
     testWrite()
     {
-        using boost::asio::buffer;
-
         doTestWrite<false>(SyncClient{});
         doTestWrite<true>(SyncClient{});
         doTestWriteDeflate(SyncClient{});
@@ -268,16 +262,46 @@ public:
         });
     }
 
+    void
+    testPausationAbandoning()
+    {
+        struct test_op
+        {
+            std::shared_ptr<stream<test::stream>> s_;
+
+            void operator()(
+                boost::system::error_code = {},
+                std::size_t = 0)
+            {
+                BEAST_FAIL();
+            }
+        };
+
+        std::weak_ptr<stream<test::stream>> weak_ws;
+        {
+            echo_server es{log};
+            net::io_context ioc;
+            auto ws = std::make_shared<stream<test::stream>>(ioc);
+            test_op op{ws};
+            ws->next_layer().connect(es.stream());
+            ws->handshake("localhost", "/");
+            ws->async_ping("", op);
+            BEAST_EXPECT(ws->impl_->wr_block.is_locked());
+            ws->async_write(sbuf("*"), op);
+            weak_ws = ws;
+            ws->next_layer().close();
+        }
+        BEAST_EXPECT(weak_ws.expired());
+    }
+
     void
     testWriteSuspend()
     {
-        using boost::asio::buffer;
-
         // suspend on ping
         doFailLoop([&](test::fail_count& fc)
         {
             echo_server es{log};
-            boost::asio::io_context ioc;
+            net::io_context ioc;
             stream<test::stream> ws{ioc, fc};
             ws.next_layer().connect(es.stream());
             ws.handshake("localhost", "/");
@@ -290,7 +314,7 @@ public:
                         BOOST_THROW_EXCEPTION(
                             system_error{ec});
                 });
-            BEAST_EXPECT(ws.wr_block_.is_locked());
+            BEAST_EXPECT(ws.impl_->wr_block.is_locked());
             BEAST_EXPECT(count == 0);
             ws.async_write(sbuf("*"),
                 [&](error_code ec, std::size_t n)
@@ -310,7 +334,7 @@ public:
         doFailLoop([&](test::fail_count& fc)
         {
             echo_server es{log};
-            boost::asio::io_context ioc;
+            net::io_context ioc;
             stream<test::stream> ws{ioc, fc};
             ws.next_layer().connect(es.stream());
             ws.handshake("localhost", "/");
@@ -323,13 +347,13 @@ public:
                         BOOST_THROW_EXCEPTION(
                             system_error{ec});
                 });
-            BEAST_EXPECT(ws.wr_block_.is_locked());
+            BEAST_EXPECT(ws.impl_->wr_block.is_locked());
             BEAST_EXPECT(count == 0);
             ws.async_write(sbuf("*"),
                 [&](error_code ec, std::size_t)
                 {
                     ++count;
-                    if(ec != boost::asio::error::operation_aborted)
+                    if(ec != net::error::operation_aborted)
                         BOOST_THROW_EXCEPTION(
                             system_error{ec});
                 });
@@ -342,7 +366,7 @@ public:
         doFailLoop([&](test::fail_count& fc)
         {
             echo_server es{log};
-            boost::asio::io_context ioc;
+            net::io_context ioc;
             stream<test::stream> ws{ioc, fc};
             ws.next_layer().connect(es.stream());
             ws.handshake("localhost", "/");
@@ -359,7 +383,7 @@ public:
                         BOOST_THROW_EXCEPTION(
                             system_error{ec});
                 });
-            while(! ws.wr_block_.is_locked())
+            while(! ws.impl_->wr_block.is_locked())
             {
                 ioc.run_one();
                 if(! BEAST_EXPECT(! ioc.stopped()))
@@ -384,7 +408,7 @@ public:
         doFailLoop([&](test::fail_count& fc)
         {
             echo_server es{log, kind::async_client};
-            boost::asio::io_context ioc;
+            net::io_context ioc;
             stream<test::stream> ws{ioc, fc};
             ws.next_layer().connect(es.stream());
             es.async_handshake();
@@ -392,7 +416,7 @@ public:
             std::size_t count = 0;
             std::string const s(16384, '*');
             ws.auto_fragment(false);
-            ws.async_write(buffer(s),
+            ws.async_write(net::buffer(s),
                 [&](error_code ec, std::size_t n)
                 {
                     ++count;
@@ -401,7 +425,7 @@ public:
                             system_error{ec});
                     BEAST_EXPECT(n == 16384);
                 });
-            BEAST_EXPECT(ws.wr_block_.is_locked());
+            BEAST_EXPECT(ws.impl_->wr_block.is_locked());
             ws.async_ping("",
                 [&](error_code ec)
                 {
@@ -418,7 +442,7 @@ public:
         doFailLoop([&](test::fail_count& fc)
         {
             echo_server es{log, kind::async_client};
-            boost::asio::io_context ioc;
+            net::io_context ioc;
             stream<test::stream> ws{ioc, fc};
             ws.next_layer().connect(es.stream());
             es.async_handshake();
@@ -426,7 +450,7 @@ public:
             std::size_t count = 0;
             std::string const s(16384, '*');
             ws.auto_fragment(true);
-            ws.async_write(buffer(s),
+            ws.async_write(net::buffer(s),
                 [&](error_code ec, std::size_t n)
                 {
                     ++count;
@@ -435,7 +459,7 @@ public:
                             system_error{ec});
                     BEAST_EXPECT(n == 16384);
                 });
-            BEAST_EXPECT(ws.wr_block_.is_locked());
+            BEAST_EXPECT(ws.impl_->wr_block.is_locked());
             ws.async_ping("",
                 [&](error_code ec)
                 {
@@ -452,14 +476,14 @@ public:
         doFailLoop([&](test::fail_count& fc)
         {
             echo_server es{log};
-            boost::asio::io_context ioc;
+            net::io_context ioc;
             stream<test::stream> ws{ioc, fc};
             ws.next_layer().connect(es.stream());
             ws.handshake("localhost", "/");
             std::size_t count = 0;
             std::string const s(16384, '*');
             ws.auto_fragment(false);
-            ws.async_write(buffer(s),
+            ws.async_write(net::buffer(s),
                 [&](error_code ec, std::size_t n)
                 {
                     ++count;
@@ -468,7 +492,7 @@ public:
                             system_error{ec});
                     BEAST_EXPECT(n == 16384);
                 });
-            BEAST_EXPECT(ws.wr_block_.is_locked());
+            BEAST_EXPECT(ws.impl_->wr_block.is_locked());
             ws.async_ping("",
                 [&](error_code ec)
                 {
@@ -484,14 +508,14 @@ public:
         doFailLoop([&](test::fail_count& fc)
         {
             echo_server es{log};
-            boost::asio::io_context ioc;
+            net::io_context ioc;
             stream<test::stream> ws{ioc, fc};
             ws.next_layer().connect(es.stream());
             ws.handshake("localhost", "/");
             std::size_t count = 0;
             std::string const s(16384, '*');
             ws.auto_fragment(true);
-            ws.async_write(buffer(s),
+            ws.async_write(net::buffer(s),
                 [&](error_code ec, std::size_t n)
                 {
                     ++count;
@@ -500,7 +524,7 @@ public:
                             system_error{ec});
                     BEAST_EXPECT(n == 16384);
                 });
-            BEAST_EXPECT(ws.wr_block_.is_locked());
+            BEAST_EXPECT(ws.impl_->wr_block.is_locked());
             ws.async_ping("",
                 [&](error_code ec)
                 {
@@ -516,7 +540,7 @@ public:
         doFailLoop([&](test::fail_count& fc)
         {
             echo_server es{log, kind::async};
-            boost::asio::io_context ioc;
+            net::io_context ioc;
             stream<test::stream> ws{ioc, fc};
             {
                 permessage_deflate pmd;
@@ -529,7 +553,7 @@ public:
             std::size_t count = 0;
             auto const& s = random_string();
             ws.binary(true);
-            ws.async_write(buffer(s),
+            ws.async_write(net::buffer(s),
                 [&](error_code ec, std::size_t n)
                 {
                     ++count;
@@ -538,7 +562,7 @@ public:
                             system_error{ec});
                     BEAST_EXPECT(n == s.size());
                 });
-            BEAST_EXPECT(ws.wr_block_.is_locked());
+            BEAST_EXPECT(ws.impl_->wr_block.is_locked());
             ws.async_ping("",
                 [&](error_code ec)
                 {
@@ -560,7 +584,7 @@ public:
             {
                 echo_server es{log, i==1 ?
                     kind::async : kind::sync};
-                boost::asio::io_context ioc;
+                net::io_context ioc;
                 stream<test::stream> ws{ioc};
                 ws.next_layer().connect(es.stream());
 
@@ -569,7 +593,7 @@ public:
                 if(! BEAST_EXPECTS(! ec, ec.message()))
                     break;
                 ws.async_write_some(false,
-                    boost::asio::const_buffer{},
+                    net::const_buffer{},
                     [&](error_code, std::size_t)
                     {
                         fail();
@@ -597,7 +621,7 @@ public:
         {
             echo_server es{log, i==1 ?
                 kind::async : kind::sync};
-            boost::asio::io_context ioc;
+            net::io_context ioc;
             stream<test::stream> ws{ioc};
             ws.next_layer().connect(es.stream());
 
@@ -613,32 +637,13 @@ public:
         }
     }
 
-    void
-    testContHook()
-    {
-        struct handler
-        {
-            void operator()(error_code) {}
-        };
-
-        char buf[32];
-        stream<test::stream> ws{ioc_};
-        stream<test::stream>::write_some_op<
-            boost::asio::const_buffer,
-                handler> op{handler{}, ws, true,
-                    boost::asio::const_buffer{
-                        buf, sizeof(buf)}};
-        using boost::asio::asio_handler_is_continuation;
-        asio_handler_is_continuation(&op);
-    }
-
     void
     testMoveOnly()
     {
-        boost::asio::io_context ioc;
+        net::io_context ioc;
         stream<test::stream> ws{ioc};
         ws.async_write_some(
-            true, boost::asio::const_buffer{},
+            true, net::const_buffer{},
             move_only_handler{});
     }
 
@@ -658,12 +663,15 @@ public:
         // breakpoint in asio_handler_invoke to make sure
         // it is instantiated.
         {
-            boost::asio::io_context ioc;
-            boost::asio::io_service::strand s{ioc};
+            net::io_context ioc;
+            net::strand<
+                net::io_context::executor_type> s(
+                    ioc.get_executor());
             stream<test::stream> ws{ioc};
             flat_buffer b;
-            ws.async_write(boost::asio::const_buffer{},
-                s.wrap(copyable_handler{}));
+            ws.async_write(net::const_buffer{},
+                net::bind_executor(s,
+                    copyable_handler{}));
         }
     }
 
@@ -671,10 +679,10 @@ public:
     run() override
     {
         testWrite();
+        testPausationAbandoning();
         testWriteSuspend();
         testAsyncWriteFrame();
         testIssue300();
-        testContHook();
         testMoveOnly();
     }
 };
